/* $Id: driver_hd44780.c,v 1.3 2005/04/30 15:07:55 dijkstra Exp $ */

/*
 * Copyright (c) 2005 Willem Dijkstra
 *                    Jorge Sanchez Pelaez beto
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if defined (HD44780)

#include <sys/types.h>

#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "driver.h"
#include "error.h"
#include "lcd.h"
#include "port.h"
#include "time.h"

__BEGIN_DECLS
void setup_block();
void setup_vbar();
void writecontrol(int);
void writedata(int);
__END_DECLS

char *driver_name = LCD_NAME;
char rowaddress[LCD_Y] = HD_ROWSTART;
int driver_capabilities = DC_BLOCK;
int data_port = 0;
int control_port = 0;

void
driver_connect(char *port, int capabilities)
{
    long portnum;
    driver_capabilities = 0;

    if (capabilities & DC_HBAR) {
	fatal("%s: cannot support horizontal bars",
	      driver_name);
    }

    portnum = strtol(port, NULL, 0);

    if (errno) {
	fatal("%s: cannot decode port number %.200s",
	      driver_name, port);
    }

    data_port = portnum;
    control_port = portnum + LPT_CONTROL;

    accessport(data_port);
    accessport(control_port);

    if (capabilities & DC_VBAR) {
	setup_block();
	setup_vbar();
    }
}
char
driver_get_hchar(int percentage)
{
    fatal("%s: horizontal bars not implemented", driver_name);

    return 0; /* NOT REACHED */
}
char
driver_get_vchar(int percentage)
{
    int unit = 1000 / 8;
    int i;

    percentage %= 101;
    percentage += 5;
    percentage *= 10;

    i = percentage / unit;

    if (i == 0) {
	return HD_CHAR_SPACE;
    }

    if (i == 8) {
	return HD_CHAR_BLOCK;
    }

    return HD_CHAR_VBAR + i - 1;
}
void
driver_nprintf(int maxsize, char *fmt, ...)
{
    char buf[LCD_MAXSTR];
    int i;
    int size;
    va_list args;

    maxsize %= sizeof(buf);

    va_start(args, fmt);
    vsnprintf(buf, maxsize, fmt, args);
    va_end(args);
    size = strlen(buf);

    for (i = 0; i < size; i++) {
	writedata(buf[i]);
    }
}
void
driver_printf(char *fmt, ...)
{
    char buf[LCD_MAXSTR];
    int i;
    int size;
    va_list args;

    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);
    size = strlen(buf);
    va_end(args);

    for (i = 0; i < size; i++) {
	writedata(buf[i]);
    }
}
void
driver_reset()
{
    writecontrol((HDCMD_FUNCTION |
		  HDCMD_FUNCTION_8BIT |
		  HDCMD_FUNCTION_2LINE |
		  HDCMD_FUNCTION_FONT5X7));
    writecontrol((HDCMD_DISPLAY_CONTROL |
		  HDCMD_DISPLAY_CONTROL_DISPLAY_OFF |
		  HDCMD_DISPLAY_CONTROL_CURSOR_OFF |
		  HDCMD_DISPLAY_CONTROL_BLINK_OFF));
    writecontrol((HDCMD_DISPLAY_CONTROL |
		  HDCMD_DISPLAY_CONTROL_DISPLAY_ON));
    writecontrol((HDCMD_ENTRY_MODE |
		  HDCMD_ENTRY_MODE_CURSOR_INCREMENT));
    writecontrol(HDCMD_CURSOR_HOME);
}
void
driver_set_backlight(int backlight)
{
    fatal("%s: backlight control not implemented", driver_name);
}
void
driver_set_contrast(int contrast)
{
    fatal("%s: contrast control not implemented", driver_name);
}
void
driver_set_specialchar(int ch, struct specialchar *data)
{
    int ch_data;
    int i, bit, mask;

    if ((ch < HD_CHAR_SPECIAL) || (ch > HD_CHAR_SPECIAL + 8)) {
	fatal("%s: cannot set specialchar %d",
	      driver_name, ch);
    }

    writecontrol(HDCMD_SETCGRAM | (ch * 8));
    for (i = 0; i < 8; i++) {
	ch_data = 0;
	mask = 0xF;

	for (bit=0; bit<6; bit++) {
	    ch_data += ((data->line[i] & mask) ? (1 << bit) : 0);
	    mask <<= 4;
	}
	writedata(ch_data);
    }
}
void
driver_set_xy(int x, int y)
{
    writecontrol(HDCMD_CURSOR_HOME);
    writecontrol((HDCMD_SETDDRAM | rowaddress[y % LCD_Y]) + (x % LCD_X));
}
void
setup_block()
{
    struct specialchar block = {
	{
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	}};

    driver_set_specialchar(HD_CHAR_BLOCK, &block);
    driver_capabilities |= DC_BLOCK;
}
void
setup_vbar()
{
    int i;
    struct specialchar vbar[] = {
	{{
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x11111,
	}},
	{{
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x11111,
	    0x11111,
	}},
	{{
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x11111,
	    0x11111,
	    0x11111,
	}},
	{{
	    0x00000,
	    0x00000,
	    0x00000,
	    0x00000,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	}},
	{{
	    0x00000,
	    0x00000,
	    0x00000,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	}},
	{{
	    0x00000,
	    0x00000,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	}},
	{{
	    0x00000,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	    0x11111,
	}}
    };

    for (i = 0; i < (sizeof(vbar) / sizeof(struct specialchar)); i++) {
	driver_set_specialchar(HD_CHAR_VBAR + i, &vbar[i]);
    }

    driver_capabilities &= !DC_HBAR;
    driver_capabilities |= DC_VBAR;
}
void
writecontrol(int data)
{
    writeport(control_port, LPT_CONTROL_MASK);
    writeport(control_port, (HDPIN_ENABLE ^ LPT_CONTROL_MASK));
    waitnsecs(40);
    writeport(data_port, data);
    waitnsecs(40);
    writeport(control_port, LPT_CONTROL_MASK);
}
void
writedata(int data)
{
    writeport(control_port, (HDPIN_RS ^ LPT_CONTROL_MASK));
    writeport(control_port, ((HDPIN_ENABLE | HDPIN_RS) ^ LPT_CONTROL_MASK));
    waitnsecs(40);
    writeport(data_port, data);
    waitnsecs(40);
    writeport(control_port, LPT_CONTROL_MASK);
}

#endif
